# njs-prometheus-module
Exposes prometheus metrics endpoint directly from N+

## Module info

prometheus-njs is an njs module written to convert miscellaneous NGINX metrics exposed by the api module to a prometheus compliant format. It uses subrequests to the /api endpoint to access the metrics.

### Module variables

You can set variables in NGINX using the `set` command. Setting variables in the same server or location block as the prometheus njs module will adjust how it processes requests. Here is a full list of variables you can set:

#### $prom_metrics_disabled
Disable metrics exporting for a particular endpoint. If you are concerned about the amount of storage NGINX metrics are taking in Prometheus, you can disable particular endpoints via this variable. Value must be comma separated list. Here is an example:
```
...
location /metrics {
    set $prom_metrics_disabled "stream/upstreams, resolvers";
    js_content prometheus.metrics;
}
...
```

A list of supported endpoints are:

* nginx
* processes
* resolvers
* connections
* ssl
* slabs
* http/requests
* http/caches
* http/upstreams
* http/server_zones
* http/location_zones
* stream/upstreams
* stream/server_zones
* stream/zone_sync
* workers/mem

#### $prom_keyval
When configuring NGINX upstreams, it is possible to use a combination of NGINX key/value store and generic upstream names to dynamically configure upstream routing without having to reload NGINX. However, this means that the upstream names inside the NGINX configuration file will be generic and not useful for metric reporting. Setting the `$prom_keyval` variable in the nginx config will allow you to specify a keyval containing a map of upstream replacements with the form `{replacement : upstream-to-replace}`. This will replace any upstreams in the prometheus output matching the `upstream-to-replace` and convert them using the corresponding `replacement`. See the provided [nginx.conf](nginx.conf#49) file for an example of how to configure `$prom_keyval`.

Here is an example of dynamic upstreams with NGINX:

```
keyval_zone zone=upstream_keyval:128k;
keyval $domain $placeholder zone=upstream_keyval;

upstream 0 {
    zone dynamic 64k;
}
upstream 1 {
    zone dynamic 64k;
}
...
upstream n {
    zone dynamic 64k;
}

server ... {
...
    location / {
        proxy_pass $scheme://$placeholder;
    }
}
```

#### $prom_keyval_stream
If you want to do dynamic replacements in the stream context, you will need to specify a new keyval in the stream zone and set the name of the keyval in the `$prom_keyval_stream` variable. See the provided [nginx.conf](nginx.conf#50) file for an example of how to configure `$prom_keyval_stream`.

```
location /metrics {
    set $prom_keyval "http_keyval";
    set $prom_keyval_stream "stream_keyval";
    js_content prometheus.metrics;
}
```

### Exported Metrics
* The following metrics are exported:
    * [NGINX Meta](http://nginx.org/en/docs/http/ngx_http_api_module.html#nginx)
    * [Processes](http://nginx.org/en/docs/http/ngx_http_api_module.html#processes)
    * [Connections](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_connections)
    * [SSL](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_ssl_object)
    * [Slabs](http://nginx.org/en/docs/http/ngx_http_api_module.html#slabs_)
    * [Resolvers](http://nginx.org/en/docs/http/ngx_http_api_module.html#resolvers_)
    * [HTTP](http://nginx.org/en/docs/http/ngx_http_api_module.html#http_)
    * [HTTP Caches](http://nginx.org/en/docs/http/ngx_http_api_module.html#http_caches_)
    * [HTTP Server Zones](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_http_server_zone)
    * [Stream Server Zones](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_stream_server_zone)
    * [HTTP Upstreams](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_http_upstream) Note: for the `state` metric, the string values are converted using the following rule: `"up"` -> `1`, `"draining"` -> `2`, `"down"` -> `3`, `"unavail"` –> `4`, `"checking"` –> `5`, `"unhealthy"` -> `6`.
    * [Stream Upstreams](http://nginx.org/en/docs/http/ngx_http_api_module.html#def_nginx_stream_upstream) Note: for the `state` metric, the string values are converted to float64 using the following rule: `"up"` -> `1`, `"draining"` -> `2`, `"down"` -> `3`, `"unavail"` –> `4`, `"checking"` –> `5`, `"unhealthy"` -> `6`
    * [Zone Sync](http://nginx.org/en/docs/http/ngx_http_api_module.html#stream_zone_sync_)
    * Worker Process Memory Usage: Aggregate Resident Set Size (RSS) and Private memory usage for all worker processes

The output of the module is Prometheus-specific metrics exported from the location block specified in the [config](nginx.conf).

### Building The Module

1. Download prometheus-exporter module sources from github
    ```
    $ git clone git@github.com:nginxinc/njs-prometheus-module.git
    ```
1. Install [NGINX Plus](https://docs.nginx.com/nginx/admin-guide/installing-nginx/installing-nginx-plus/)

1. Include the `prometheus.js` file inside the nginx.conf file.
    ```
    js_import prometheus.js;
    ```
1. Expose api metrics at the endpoint /metrics
    ```
    location /api {
        api write=on;
    }
    ```
1. Expose an endpoint for prometheus metrics
    ```
    location /metrics {
        js_content prometheus.metrics;
    }
    ```
1. [Optional] set the `$prom_keyval` variable with the keyval holding upstream replacements
    ```
    set $prom_keyval "upstream_keyval";
    ```
    Note: You will need to fill the `upstream_keyval` keyval with a map of the form `{replacement : upstream-to-replace}`.


An example NGINX configuration file is provided [here](nginx.conf). To replace the upstream `0` with a more informative upstream name, add to the upstream_keyvals map:
```
curl localhost/api/4/http/keyvals/upstream_keyval -X POST -d '{"backend":"0"}'
```
